/*
Copyright (C) THL A29 Limited, a Tencent company. All rights reserved.

SPDX-License-Identifier: Apache-2.0
*/

package tbft

import (
	"crypto/rand"
	"crypto/sha256"
	"testing"
	"time"

	"chainmaker.org/chainmaker/common/v2/msgbus"
	"chainmaker.org/chainmaker/consensus-utils/v2/testframework"
	"chainmaker.org/chainmaker/logger/v2"
	commonpb "chainmaker.org/chainmaker/pb-go/v2/common"
	"github.com/golang/mock/gomock"

	tbftpb "chainmaker.org/chainmaker/pb-go/v2/consensus/tbft"

	configpb "chainmaker.org/chainmaker/pb-go/v2/config"

	consensuspb "chainmaker.org/chainmaker/pb-go/v2/consensus"

	"chainmaker.org/chainmaker/consensus-utils/v2/consistent_service"
	"chainmaker.org/chainmaker/protocol/v2/mock"

	"github.com/stretchr/testify/require"
)

//
//  TestConsensusTBFTImpl_Type
//  @Description: TestConsensusTBFTImpl_Type
//  @param t
//
func TestConsensusTBFTImpl_Type(t *testing.T) {
	tbftImpl.RLock()
	defer tbftImpl.RUnlock()
	typeInt := tbftImpl.Type()
	require.Equal(t, typeInt, int8(TypeLocalTBFTState))
}

//
//  TestConsensusTBFTImpl_Data
//  @Description: TestConsensusTBFTImpl_Data
//  @param t
//
func TestConsensusTBFTImpl_Data(t *testing.T) {
	tbftImpl.RLock()
	defer tbftImpl.RUnlock()
	date := tbftImpl.Data()
	require.Equal(t, date, tbftImpl)
}

//
//  TestConsensusTBFTImpl_Update
//  @Description: TestConsensusTBFTImpl_Update
//  @param t
//
func TestConsensusTBFTImpl_Update(t *testing.T) {
	tbftImpl.Lock()
	defer tbftImpl.Unlock()
	status := make(map[int8]consistent_service.Status)
	s := &RemoteState{Id: "node1", Height: 1, Round: 1, Step: 1}
	status[s.Type()] = s
	tbftImpl.Update(s)

}

//
//  TestConsensusTBFTImpl_InitExtendHandler
//  @Description: TestConsensusTBFTImpl_InitExtendHandler
//  @param t
//
func TestConsensusTBFTImpl_InitExtendHandler(t *testing.T) {
	tbftImpl.Lock()
	defer tbftImpl.Unlock()
	handler := &mock.MockConsensusExtendHandler{}
	tbftImpl.InitExtendHandler(handler)
}

//
//  TestConsensusTBFTImpl_Verify
//  @Description: TestConsensusTBFTImpl_Verify
//  @param t
//
func TestConsensusTBFTImpl_Verify(t *testing.T) {
	tbftImpl.Lock()
	defer tbftImpl.Unlock()
	chainConfig := &configpb.ChainConfig{
		Consensus: &configpb.ConsensusConfig{
			Type: consensuspb.ConsensusType_TBFT,
			Nodes: []*configpb.OrgConfig{
				{
					OrgId:  org1Id,
					NodeId: []string{org1NodeId},
				},
				{
					OrgId:  org2Id,
					NodeId: []string{org2NodeId},
				},
				{
					OrgId:  org3Id,
					NodeId: []string{org3NodeId},
				},
				{
					OrgId:  org4Id,
					NodeId: []string{org4NodeId},
				},
			},
		},
	}
	err := tbftImpl.Verify(consensuspb.ConsensusType_SOLO, chainConfig)
	require.NotNil(t, err)
}

//
//  TestConsensusTBFTImpl_ExtractProposeTimeout
//  @Description: TestConsensusTBFTImpl_ExtractProposeTimeout
//  @param t
//
func TestConsensusTBFTImpl_ExtractProposeTimeout(t *testing.T) {
	tbftImpl.Lock()
	defer tbftImpl.Unlock()

	_, err := tbftImpl.extractProposeTimeout("100")
	require.NotNil(t, err)

	_, err = tbftImpl.extractProposeTimeout((time.Millisecond * 100).String())
	require.Nil(t, err)
}

//
//  TestConsensusTBFTImpl_ExtractProposeTimeoutDelta
//  @Description: TestConsensusTBFTImpl_ExtractProposeTimeoutDelta
//  @param t
//
func TestConsensusTBFTImpl_ExtractProposeTimeoutDelta(t *testing.T) {
	tbftImpl.Lock()
	defer tbftImpl.Unlock()

	_, err := tbftImpl.extractProposeTimeoutDelta("100")
	require.NotNil(t, err)

	_, err = tbftImpl.extractProposeTimeoutDelta((time.Millisecond * 100).String())
	require.Nil(t, err)
}

//
//  TestConsensusTBFTImpl_ExtractBlocksPerProposer
//  @Description: TestConsensusTBFTImpl_ExtractBlocksPerProposer
//  @param t
//
func TestConsensusTBFTImpl_ExtractBlocksPerProposer(t *testing.T) {
	tbftImpl.Lock()
	defer tbftImpl.Unlock()

	_, err := tbftImpl.extractBlocksPerProposer("100")
	require.Nil(t, err)

	_, err = tbftImpl.extractBlocksPerProposer((time.Millisecond * 100).String())
	require.NotNil(t, err)

	_, err = tbftImpl.extractBlocksPerProposer("-1")
	require.NotNil(t, err)
}

//
//  TestConsensusTBFTImpl_handleTimeout
//  @Description: TestConsensusTBFTImpl_handleTimeout
//  @param t
//
func TestConsensusTBFTImpl_handleTimeout(t *testing.T) {

	timeoutInfo := &tbftpb.TimeoutInfo{Duration: 1, Height: 1, Round: 1, Step: -1}
	tbftImpl.handleTimeout(*timeoutInfo)

	timeoutInfo = &tbftpb.TimeoutInfo{Duration: 1, Height: 1, Round: 1, Step: tbftpb.Step_PREVOTE}
	tbftImpl.handleTimeout(*timeoutInfo)

	timeoutInfo = &tbftpb.TimeoutInfo{Duration: 1, Height: 1, Round: 1, Step: tbftpb.Step_PRECOMMIT}
	tbftImpl.handleTimeout(*timeoutInfo)

	timeoutInfo = &tbftpb.TimeoutInfo{Duration: 1, Height: 1, Round: 1, Step: tbftpb.Step_COMMIT}
	tbftImpl.handleTimeout(*timeoutInfo)
}

//
//  TestConsensusTBFTImpl_PrevoteTimeout
//  @Description: TestConsensusTBFTImpl_PrevoteTimeout
//  @param t
//
func TestConsensusTBFTImpl_PrevoteTimeout(t *testing.T) {
	t1 := time.Duration(
		TimeoutPrevote.Nanoseconds()+TimeoutPrevoteDelta.Nanoseconds()*int64(0),
	) * time.Nanosecond
	t2 := tbftImpl.PrevoteTimeout(0)
	require.Equal(t, t1, t2)
}

//
//  TestConsensusTBFTImpl_PrecommitTimeout
//  @Description: TestConsensusTBFTImpl_PrecommitTimeout
//  @param t
//
func TestConsensusTBFTImpl_PrecommitTimeout(t *testing.T) {
	t1 := time.Duration(
		TimeoutPrecommit.Nanoseconds()+TimeoutPrecommitDelta.Nanoseconds()*int64(0),
	) * time.Nanosecond
	t2 := tbftImpl.PrecommitTimeout(0)
	require.Equal(t, t1, t2)
}

//
//  TestConsensusTBFTImpl_CommitTimeout
//  @Description: TestConsensusTBFTImpl_CommitTimeout
//  @param t
//
func TestConsensusTBFTImpl_CommitTimeout(t *testing.T) {
	t1 := time.Duration(TimeoutCommit.Nanoseconds()*int64(0)) * time.Nanosecond
	t2 := tbftImpl.CommitTimeout(0)
	require.Equal(t, t1, t2)
}

//
// TestConsensusTBFTImpl_ToProto
// @Description: TestConsensusTBFTImpl_ToProto
// @param t
//
func TestConsensusTBFTImpl_ToProto(t *testing.T) {
	tbftImpl.RLock()
	defer tbftImpl.RUnlock()
	consensusState := tbftImpl.ToProto()
	require.NotNil(t, consensusState)
}

//
// TestConsensusTBFTImpl_ToGossipStateProto
// @Description: TestConsensusTBFTImpl_ToGossipStateProto
// @param t
//
func TestConsensusTBFTImpl_ToGossipStateProto(t *testing.T) {
	tbftImpl.RLock()
	defer tbftImpl.RUnlock()
	gssipState := tbftImpl.ToGossipStateProto()
	require.NotNil(t, gssipState)
}

//
// TestConsensusTBFTImpl_EnterPrecommitFromReplayWal
// @Description: TestConsensusTBFTImpl_EnterPrecommitFromReplayWal
// @param t
//
func TestConsensusTBFTImpl_EnterPrecommitFromReplayWal(t *testing.T) {
	tbftImpl.RLock()
	defer tbftImpl.RUnlock()
	err := tbftImpl.enterPrecommitFromReplayWal(nil)
	require.NotNil(t, err)
}

//
// TestConsensusTBFTImpl_GetValidators
// @Description: TestConsensusTBFTImpl_GetValidators
// @param t
//
func TestConsensusTBFTImpl_GetValidators(t *testing.T) {
	tbftImpl.RLock()
	defer tbftImpl.RUnlock()
	_, err := tbftImpl.GetValidators()
	require.Nil(t, err)
}

//
// TestConsensusTBFTImpl_GetLastHeight
// @Description: TestConsensusTBFTImpl_GetLastHeight
// @param t
//
func TestConsensusTBFTImpl_GetLastHeight(t *testing.T) {
	tbftImpl.RLock()
	defer tbftImpl.RUnlock()
	height := tbftImpl.GetLastHeight()
	require.NotNil(t, height)
}

//
// TestConsensusTBFTImpl_GetConsensusStateJSON
// @Description: TestConsensusTBFTImpl_GetConsensusStateJSON
// @param t
//
func TestConsensusTBFTImpl_GetConsensusStateJSON(t *testing.T) {
	tbftImpl.RLock()
	defer tbftImpl.RUnlock()
	_, err := tbftImpl.GetConsensusStateJSON()
	require.Nil(t, err)
}

func TestConsensusTBFTImpl_getValidatorSet(t *testing.T) {
	vs := tbftImpl.getValidatorSet()
	require.NotNil(t, vs)
}

func TestConsensusTBFTImpl_onQuit(t *testing.T) {
	tbftImpl.OnQuit()
}

//
//  TestProcVerifyBlockWithRwSets
//  @Description: TestProcVerifyBlockWithRwSets
//  @param t
//
func TestProcVerifyBlockWithRwSets(t *testing.T) {
	ctrl := gomock.NewController(t)
	defer ctrl.Finish()

	tbftImpl.RLock()
	defer tbftImpl.RUnlock()
	var blockHeight uint64 = 10
	blockHash := sha256.Sum256(nil)
	_, _ = rand.Read(blockHash[:])
	block := &commonpb.Block{
		Header: &commonpb.BlockHeader{
			BlockHeight: blockHeight,
			BlockHash:   blockHash[:],
		},
		AdditionalData: &commonpb.AdditionalData{
			ExtraData: map[string][]byte{
				TBFTAddtionalDataKey: nil,
			},
		},
	}

	var votes []*tbftpb.Vote
	for _, id := range tbftImpl.validatorSet.Validators {
		vote := NewVote(tbftpb.VoteType_VOTE_PREVOTE, id, blockHeight, 0, blockHash[:])
		votes = append(votes, vote)
	}

	proposal := &tbftpb.Proposal{
		Voter:    "node1",
		Height:   1,
		Round:    1,
		PolRound: 1,
		Block:    block,
		Qc:       votes,
	}

	tbftProposal := NewTBFTProposal(proposal, true)
	oldLen := len(tbftProposal.Bytes)
	tbftProposal.Marshal()
	require.Equal(t, oldLen, len(tbftProposal.Bytes))
	core := testframework.NewCoreEngineForTest(&testframework.TestNodeConfig{}, cmLogger)
	ac := mock.NewMockAccessControlProvider(ctrl)
	ac.EXPECT().CreatePrincipal(gomock.Any(), gomock.Any(), gomock.Any()).AnyTimes().Return(nil, nil)
	ac.EXPECT().VerifyPrincipal(gomock.Any()).AnyTimes().Return(true, nil)
	tbftImpl.ac = ac
	tbftImpl.blockVerifier = core.GetBlockVerifier()
	err := tbftImpl.procVerifyBlockWithRwSets(proposal)
	require.Nil(t, err)
}

//
//  TestProcRoundQc
//  @Description: TestProcRoundQc
//  @param t
//
func TestProcRoundQc(t *testing.T) {
	ctrl := gomock.NewController(t)
	defer ctrl.Finish()

	tbftImpl.RLock()
	defer tbftImpl.RUnlock()

	tbftImpl.heightRoundVoteSet = newHeightRoundVoteSet(
		cmLogger, 1, 1, tbftImpl.validatorSet)

	blockHash := sha256.Sum256(nil)
	_, _ = rand.Read(blockHash[:])
	for _, id := range tbftImpl.validatorSet.Validators {
		vote := NewVote(tbftpb.VoteType_VOTE_PRECOMMIT, id, 1, 1, blockHash[:])
		tbftImpl.heightRoundVoteSet.addVote(vote)
	}

	voteSet := NewVoteSet(cmLogger, tbftpb.VoteType_VOTE_PRECOMMIT, 1, 2, tbftImpl.validatorSet)

	for _, id := range tbftImpl.validatorSet.Validators {
		vote := NewVote(tbftpb.VoteType_VOTE_PRECOMMIT, id, 1, 2, blockHash[:])
		_, _ = voteSet.AddVote(vote, false)
	}

	roundqc := &tbftpb.RoundQC{
		Height: 1,
		Round:  2,
		Qc:     voteSet.ToProto(),
	}

	ac := mock.NewMockAccessControlProvider(ctrl)
	ac.EXPECT().CreatePrincipal(gomock.Any(), gomock.Any(), gomock.Any()).AnyTimes().Return(nil, nil)
	ac.EXPECT().VerifyPrincipal(gomock.Any()).AnyTimes().Return(true, nil)
	tbftImpl.ac = ac
	tbftImpl.procRoundQC(roundqc)
}

//
//  TestProcRoundQcByPC
//  @Description: TestProcRoundQcByPC
//  @param t
//
func TestProcRoundQcByPC(t *testing.T) {
	ctrl := gomock.NewController(t)
	defer ctrl.Finish()

	tbftImpl.RLock()
	defer tbftImpl.RUnlock()

	tbftImpl.metrics = newHeightMetrics(1)
	tbftImpl.heightRoundVoteSet = newHeightRoundVoteSet(
		cmLogger, 1, 1, tbftImpl.validatorSet)

	blockHash := sha256.Sum256(nil)
	_, _ = rand.Read(blockHash[:])

	voteSet := NewVoteSet(cmLogger, tbftpb.VoteType_VOTE_PREVOTE, 1, 2, tbftImpl.validatorSet)

	for _, id := range tbftImpl.validatorSet.Validators {
		vote := NewVote(tbftpb.VoteType_VOTE_PRECOMMIT, id, 1, 2, blockHash[:])
		_, _ = voteSet.AddVote(vote, false)
	}

	roundqc := &tbftpb.RoundQC{
		Height: 1,
		Round:  2,
		Qc:     voteSet.ToProto(),
	}

	ac := mock.NewMockAccessControlProvider(ctrl)
	ac.EXPECT().CreatePrincipal(gomock.Any(), gomock.Any(), gomock.Any()).AnyTimes().Return(nil, nil)
	ac.EXPECT().VerifyPrincipal(gomock.Any()).AnyTimes().Return(true, nil)
	tbftImpl.ac = ac
	tbftImpl.procRoundQC(roundqc)
}

//
//  TestDelInvalidTxs
//  @Description: TestDelInvalidTxs
//  @param t
//
func TestDelInvalidTxs(t *testing.T) {
	ctrl := gomock.NewController(t)
	defer ctrl.Finish()

	tbftImpl.RLock()
	defer tbftImpl.RUnlock()

	//blockHash := sha256.Sum256(nil)
	//_, _ = rand.Read(blockHash[:])

	voteSet := NewVoteSet(cmLogger, tbftpb.VoteType_VOTE_PREVOTE, 1, 2, tbftImpl.validatorSet)

	for _, id := range tbftImpl.validatorSet.Validators {
		vote := NewVote(tbftpb.VoteType_VOTE_PREVOTE, id, 1, 2, nilHash)
		vote.InvalidTxs = []string{"hash1", "hash2"}
		_, _ = voteSet.AddVote(vote, true)
	}

	tbftImpl.delInvalidTxs(voteSet, nilHash)
}

func TestTBFTConsensus_OnQuit(t *testing.T) {
	type fields struct {
		log      *logger.CMLogger
		msgBus   msgbus.MessageBus
		messageC chan interface{}
	}
	tests := []struct {
		name   string
		fields fields
	}{
		{"test1",
			fields{logger.GetLogger(chainid), tbftImpl.msgbus, make(chan interface{})},
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			m := &ConsensusTBFTImpl{
				logger: tt.fields.log,
				msgbus: tt.fields.msgBus,
			}
			m.OnQuit()
		})
	}
}
