package tbft

import (
	"crypto/rand"
	"crypto/sha256"
	"reflect"
	"testing"

	msgbus "chainmaker.org/chainmaker/common/v2/msgbus"
	"chainmaker.org/chainmaker/consensus-utils/v2/consistent_service"
	commonpb "chainmaker.org/chainmaker/pb-go/v2/common"
	tbftpb "chainmaker.org/chainmaker/pb-go/v2/consensus/tbft"
	netpb "chainmaker.org/chainmaker/pb-go/v2/net"
	"chainmaker.org/chainmaker/protocol/v2"
	"github.com/stretchr/testify/require"
)

//
//  newTbftImpl
//  @Description: new tbft impl
//  @return *ConsensusTBFTImpl
//
func newTbftImpl() *ConsensusTBFTImpl {
	validators := []string{"validator0", "validator1"}
	valSet := &validatorSet{
		logger:           tLogger,
		Validators:       validators,
		validatorsHeight: make(map[string]uint64),
	}

	validatorSet := newValidatorSet(tLogger, nodes, 1)
	heightRoundVoteSet := newHeightRoundVoteSet(
		tLogger, 1, 0, validatorSet)
	voteHash := []byte("voteHash")
	prevoteVote := NewVote(tbftpb.VoteType_VOTE_PREVOTE, validators[0], 1, 0, voteHash[:])
	precommitVote := NewVote(tbftpb.VoteType_VOTE_PRECOMMIT, validators[0], 1, 0, voteHash[:])
	_, _ = heightRoundVoteSet.addVote(prevoteVote)
	_, _ = heightRoundVoteSet.addVote(precommitVote)
	var blockHeight uint64 = 10
	blockHash := sha256.Sum256(nil)
	_, _ = rand.Read(blockHash[:])
	block := &commonpb.Block{
		Header: &commonpb.BlockHeader{
			BlockHeight: blockHeight,
			BlockHash:   blockHash[:],
		},
		AdditionalData: &commonpb.AdditionalData{
			ExtraData: map[string][]byte{
				TBFTAddtionalDataKey: nil,
			},
		},
	}
	proposal := &tbftpb.Proposal{
		Voter:    "node1",
		Height:   1,
		Round:    1,
		PolRound: 1,
		Block:    block,
	}
	tbftProposal := NewTBFTProposal(proposal, true)
	consensusState := &ConsensusState{tLogger,
		"node1", 1, 1, 1, tbftProposal,
		tbftProposal, 1, nil, 1,
		nil, heightRoundVoteSet}
	return &ConsensusTBFTImpl{
		logger:              tLogger,
		Id:                  chainid,
		msgbus:              msgbus.NewMessageBus(),
		validatorSet:        valSet,
		ConsensusState:      consensusState,
		consensusStateCache: newConsensusStateCache(defaultConsensusStateCacheSize),
	}
}

var (
	tLogger = newMockLogger()
	chainid = "chainId1"
	sb      = &StatusBroadcaster{
		log: newMockLogger(),
		id:  StatusBroadcasterTbft,
		// 运行状态
		running: false,
	}
	tbftImpl = newTbftImpl()
	nodes    = []string{
		"QmQZn3pZCcuEf34FSvucqkvVJEvfzpNjQTk17HS6CYMR35",
		"QmeRZz3AjhzydkzpiuuSAtmqt8mU8XcRH2hynQN4tLgYg6",
		"QmTSMcqwp4X6oPP5WrNpsMpotQMSGcxVshkGLJUhCrqGbu",
		"QmUryDgjNoxfMXHdDRFZ5Pe55R1vxTPA3ZgCteHze2ET27",
	}
)

//
//  TestNewTBFTStatusBroadcaster
//  @Description: TestNewTBFTStatusBroadcaster
//  @param t
//
func TestNewTBFTStatusBroadcaster(t *testing.T) {
	type args struct {
		log protocol.Logger
	}
	tests := []struct {
		name string
		args args
		want *StatusBroadcaster
	}{
		{"test",
			args{newMockLogger()},
			sb,
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			if got := NewTBFTStatusBroadcaster(tt.args.log); !reflect.DeepEqual(got, tt.want) {
				t.Errorf("NewTBFTStatusBroadcaster() = %v, want %v", got, tt.want)
			}
		})
	}
}

//
//  TestStatusBroadcaster_ID
//  @Description: TestStatusBroadcaster_ID
//  @param t
//
func TestStatusBroadcaster_ID(t *testing.T) {
	type fields struct {
		log     protocol.Logger
		id      string
		running bool
	}
	tests := []struct {
		name   string
		fields fields
		want   string
	}{
		{"test",
			fields{sb.log, sb.id, sb.running},
			StatusBroadcasterTbft},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			tsb := &StatusBroadcaster{
				log:     tt.fields.log,
				id:      tt.fields.id,
				running: tt.fields.running,
			}
			if got := tsb.ID(); got != tt.want {
				t.Errorf("ID() = %v, want %v", got, tt.want)
			}
		})
	}
}

//
//  TestStatusBroadcaster_IsRunning
//  @Description: TestStatusBroadcaster_IsRunning
//  @param t
//
func TestStatusBroadcaster_IsRunning(t *testing.T) {
	type fields struct {
		log     protocol.Logger
		id      string
		running bool
	}
	tests := []struct {
		name   string
		fields fields
		want   bool
	}{
		{"test",
			fields{sb.log, sb.id, sb.running},
			false},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			tsb := &StatusBroadcaster{
				log:     tt.fields.log,
				id:      tt.fields.id,
				running: tt.fields.running,
			}
			if got := tsb.IsRunning(); got != tt.want {
				t.Errorf("IsRunning() = %v, want %v", got, tt.want)
			}
		})
	}
}

//
//  TestStatusBroadcaster_Run
//  @Description: TestStatusBroadcaster_Run
//  @param t
//
func TestStatusBroadcaster_Run(t *testing.T) {
	type fields struct {
		log     protocol.Logger
		id      string
		running bool
	}
	tests := []struct {
		name   string
		fields fields
	}{
		{"test",
			fields{newMockLogger(), StatusBroadcasterTbft, false},
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			tsb := &StatusBroadcaster{
				log:     tt.fields.log,
				id:      tt.fields.id,
				running: tt.fields.running,
			}
			_ = tsb.Start()
		})
	}
}

//
//  TestStatusBroadcaster_Stop
//  @Description: TestStatusBroadcaster_Stop
//  @param t
//
func TestStatusBroadcaster_Stop(t *testing.T) {
	type fields struct {
		log     protocol.Logger
		id      string
		running bool
	}
	tests := []struct {
		name   string
		fields fields
	}{
		{"test",
			fields{newMockLogger(), StatusBroadcasterTbft, false},
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			tsb := &StatusBroadcaster{
				log:     tt.fields.log,
				id:      tt.fields.id,
				running: tt.fields.running,
			}
			_ = tsb.Stop()
		})
	}
}

//
//  TestStatusBroadcaster_TimePattern
//  @Description: TestStatusBroadcaster_TimePattern
//  @param t
//
func TestStatusBroadcaster_TimePattern(t *testing.T) {
	type fields struct {
		log     protocol.Logger
		id      string
		running bool
	}
	tests := []struct {
		name   string
		fields fields
		want   interface{}
	}{
		{"test",
			fields{newMockLogger(), StatusBroadcasterTbft, false},
			TimerInterval,
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			tsb := &StatusBroadcaster{
				log:     tt.fields.log,
				id:      tt.fields.id,
				running: tt.fields.running,
			}
			if got := tsb.TimePattern(); !reflect.DeepEqual(got, tt.want) {
				t.Errorf("TimePattern() = %v, want %v", got, tt.want)
			}
		})
	}
}

//
//  TestStatusInterceptor_Handle
//  @Description: TestStatusInterceptor_Handle
//  @param t
//
func TestStatusInterceptor_Handle(t *testing.T) {
	type args struct {
		status consistent_service.Status
	}

	s1 := &RemoteState{Id: "node1"}

	tests := []struct {
		name    string
		args    args
		wantErr bool
	}{
		{"RemoteState",
			args{s1},
			false,
		},
		{"tbftImpl",
			args{tbftImpl},
			false,
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			tsb := &StatusInterceptor{}
			if err := tsb.Handle(tt.args.status); (err != nil) != tt.wantErr {
				t.Errorf("Handle() error = %v, wantErr %v", err, tt.wantErr)
			}
		})
	}
}

//
//  Test_sendPrecommitInState
//  @Description: Test_sendPrecommitInState
//  @param t
//
func Test_sendPrecommitInState(t *testing.T) {
	tbftImpl.Lock()
	defer tbftImpl.Unlock()

	var netMSGs []*netpb.NetMsg
	prevotes := NewVoteSet(tLogger, tbftpb.VoteType_VOTE_PREVOTE, 1, 0, tbftImpl.validatorSet)
	precommits := NewVoteSet(tLogger, tbftpb.VoteType_VOTE_PRECOMMIT, 1, 0, tbftImpl.validatorSet)
	voteHash := []byte("voteHash")
	prevoteVote := NewVote(tbftpb.VoteType_VOTE_PREVOTE, tbftImpl.validatorSet.Validators[0], 1, 0, voteHash[:])
	precommitVote := NewVote(tbftpb.VoteType_VOTE_PRECOMMIT, tbftImpl.validatorSet.Validators[0], 1, 0, voteHash[:])
	_, _ = prevotes.AddVote(prevoteVote, false)
	_, _ = precommits.AddVote(precommitVote, false)

	remoter := &RemoteState{
		Height:       1,
		Round:        0,
		RoundVoteSet: newRoundVoteSet(1, 0, prevotes, precommits),
	}

	//remoter := &RemoteState{}

	type args struct {
		local   *ConsensusTBFTImpl
		state   *ConsensusState
		remoter *RemoteState
		netMSGs *[]*netpb.NetMsg
	}

	tests := []struct {
		name string
		args args
	}{
		{"test",
			args{tbftImpl, tbftImpl.ConsensusState,
				remoter, &netMSGs}},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			sendPrecommitInState(tt.args.local, tt.args.state, tt.args.remoter, tt.args.netMSGs)
		})
	}
}

//
//  Test_sendPrecommitOfRound
//  @Description: Test_sendPrecommitOfRound
//  @param t
//
func Test_sendPrecommitOfRound(t *testing.T) {
	tbftImpl.Lock()
	defer tbftImpl.Unlock()

	var netMSGs []*netpb.NetMsg
	prevotes := NewVoteSet(tLogger, tbftpb.VoteType_VOTE_PREVOTE, 1, 0, tbftImpl.validatorSet)
	precommits := NewVoteSet(tLogger, tbftpb.VoteType_VOTE_PRECOMMIT, 1, 0, tbftImpl.validatorSet)
	voteHash := []byte("voteHash")
	prevoteVote := NewVote(tbftpb.VoteType_VOTE_PREVOTE, tbftImpl.validatorSet.Validators[0], 1, 0, voteHash[:])
	precommitVote := NewVote(tbftpb.VoteType_VOTE_PRECOMMIT, tbftImpl.validatorSet.Validators[0], 1, 0, voteHash[:])
	_, _ = prevotes.AddVote(prevoteVote, false)
	_, _ = precommits.AddVote(precommitVote, false)

	remoter := &RemoteState{
		Height:       1,
		Round:        0,
		RoundVoteSet: newRoundVoteSet(1, 0, prevotes, precommits),
	}
	//remoter := &RemoteState{}
	status := make(map[int8]consistent_service.Status)
	status[remoter.Type()] = remoter
	r := &Node{id: "node1", status: status}
	type args struct {
		local      *ConsensusTBFTImpl
		remoteInfo consistent_service.Node
		remoter    *RemoteState
		netMSGs    *[]*netpb.NetMsg
	}
	tests := []struct {
		name string
		args args
	}{
		{"test",
			args{tbftImpl, r,
				remoter, &netMSGs}},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			sendPrecommitOfRound(tt.args.local, tt.args.remoteInfo, tt.args.remoter, tt.args.netMSGs)
		})
	}
}

//
// Test_sendRoundQC
// @Description: Test_sendRoundQC
// @param t
//
func Test_sendRoundQC(t *testing.T) {
	tbftImpl.Lock()
	defer tbftImpl.Unlock()

	var netMSGs []*netpb.NetMsg

	tbftImpl.Height = 1
	tbftImpl.Round = 4
	//precommits := NewVoteSet(tLogger, tbftpb.VoteType_VOTE_PRECOMMIT, 1, 4, tbftImpl.validatorSet)
	voteHash := []byte("voteHash")
	precommitVote0 := NewVote(tbftpb.VoteType_VOTE_PRECOMMIT, tbftImpl.validatorSet.Validators[0], 1, 4, voteHash[:])
	precommitVote1 := NewVote(tbftpb.VoteType_VOTE_PRECOMMIT, tbftImpl.validatorSet.Validators[1], 1, 4, voteHash[:])
	_, _ = tbftImpl.heightRoundVoteSet.addVote(precommitVote0)
	_, _ = tbftImpl.heightRoundVoteSet.addVote(precommitVote1)

	remoter := &RemoteState{
		Height: 1,
		Round:  0,
	}
	//remoter := &RemoteState{}
	status := make(map[int8]consistent_service.Status)
	status[remoter.Type()] = remoter
	r := &Node{id: "node1", status: status}
	type args struct {
		local      *ConsensusTBFTImpl
		remoteInfo consistent_service.Node
		remoter    *RemoteState
		netMSGs    *[]*netpb.NetMsg
	}
	tests := []struct {
		name string
		args args
	}{
		{"test",
			args{tbftImpl, r,
				remoter, &netMSGs}},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			sendRoundQC(tt.args.local, tt.args.remoteInfo, tt.args.remoter, tt.args.netMSGs)
		})
	}
}

//
//  Test_sendPrevoteInState
//  @Description: Test_sendPrevoteInState
//  @param t
//
func Test_sendPrevoteInState(t *testing.T) {
	tbftImpl.Lock()
	defer tbftImpl.Unlock()

	var netMSGs []*netpb.NetMsg
	prevotes := NewVoteSet(tLogger, tbftpb.VoteType_VOTE_PREVOTE, 1, 0, tbftImpl.validatorSet)
	precommits := NewVoteSet(tLogger, tbftpb.VoteType_VOTE_PRECOMMIT, 1, 0, tbftImpl.validatorSet)
	voteHash := []byte("voteHash")
	prevoteVote := NewVote(tbftpb.VoteType_VOTE_PREVOTE, tbftImpl.validatorSet.Validators[0], 1, 0, voteHash[:])
	precommitVote := NewVote(tbftpb.VoteType_VOTE_PRECOMMIT, tbftImpl.validatorSet.Validators[0], 1, 0, voteHash[:])
	_, _ = prevotes.AddVote(prevoteVote, false)
	_, _ = precommits.AddVote(precommitVote, false)

	remoter := &RemoteState{
		Height:       1,
		Round:        0,
		RoundVoteSet: newRoundVoteSet(1, 0, prevotes, precommits),
	}

	//remoter := &RemoteState{}

	type args struct {
		local   *ConsensusTBFTImpl
		state   *ConsensusState
		remoter *RemoteState
		netMSGs *[]*netpb.NetMsg
	}
	tests := []struct {
		name string
		args args
	}{
		{"test",
			args{tbftImpl, tbftImpl.ConsensusState,
				remoter, &netMSGs}},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			sendPrevoteInState(tt.args.local, tt.args.state, tt.args.remoter, tt.args.netMSGs)
		})
	}
}

//
//  Test_sendPrevoteOfRound
//  @Description: Test_sendPrevoteOfRound
//  @param t
//
func Test_sendPrevoteOfRound(t *testing.T) {
	tbftImpl.Lock()
	defer tbftImpl.Unlock()

	var netMSGs []*netpb.NetMsg
	prevotes := NewVoteSet(tLogger, tbftpb.VoteType_VOTE_PREVOTE, 1, 0, tbftImpl.validatorSet)
	precommits := NewVoteSet(tLogger, tbftpb.VoteType_VOTE_PRECOMMIT, 1, 0, tbftImpl.validatorSet)
	voteHash := []byte("voteHash")
	prevoteVote := NewVote(tbftpb.VoteType_VOTE_PREVOTE, tbftImpl.validatorSet.Validators[0], 1, 0, voteHash[:])
	precommitVote := NewVote(tbftpb.VoteType_VOTE_PRECOMMIT, tbftImpl.validatorSet.Validators[0], 1, 0, voteHash[:])
	_, _ = prevotes.AddVote(prevoteVote, false)
	_, _ = precommits.AddVote(precommitVote, false)

	remoter := &RemoteState{
		Height:       1,
		Round:        0,
		RoundVoteSet: newRoundVoteSet(1, 0, prevotes, precommits),
	}
	//remoter := &RemoteState{}
	status := make(map[int8]consistent_service.Status)
	status[remoter.Type()] = remoter
	r := &Node{id: "node1", status: status}

	type args struct {
		local      *ConsensusTBFTImpl
		remoteInfo consistent_service.Node
		remoter    *RemoteState
		netMSGs    *[]*netpb.NetMsg
	}
	tests := []struct {
		name string
		args args
	}{
		{"test",
			args{tbftImpl, r,
				remoter, &netMSGs}},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			sendPrevoteOfRound(tt.args.local, tt.args.remoteInfo, tt.args.remoter, tt.args.netMSGs)
		})
	}
}

//
//  Test_sendProposalInState
//  @Description: Test_sendProposalInState
//  @param t
//
func Test_sendProposalInState(t *testing.T) {
	tbftImpl.Lock()
	defer tbftImpl.Unlock()

	var netMSGs []*netpb.NetMsg
	remoter := &RemoteState{}

	type args struct {
		state   *ConsensusState
		remoter *RemoteState
		netMSGs *[]*netpb.NetMsg
	}
	tests := []struct {
		name string
		args args
	}{
		{"test",
			args{tbftImpl.ConsensusState,
				remoter, &netMSGs}},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			sendProposalInState(tt.args.state, tt.args.remoter, tt.args.netMSGs)
		})
	}
}

//
//  Test_sendProposalOfRound
//  @Description: Test_sendProposalOfRound
//  @param t
//
func Test_sendProposalOfRound(t *testing.T) {
	tbftImpl.Lock()
	defer tbftImpl.Unlock()

	var netMSGs []*netpb.NetMsg
	remoter := &RemoteState{}
	status := make(map[int8]consistent_service.Status)
	status[remoter.Type()] = remoter
	r := &Node{id: "node1", status: status}

	type args struct {
		local      *ConsensusTBFTImpl
		remoteInfo consistent_service.Node
		remoter    *RemoteState
		netMSGs    *[]*netpb.NetMsg
	}
	tests := []struct {
		name string
		args args
	}{
		{"test",
			args{tbftImpl, r,
				remoter,
				&netMSGs}},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			sendProposalOfRound(tt.args.local, tt.args.remoteInfo, tt.args.remoter, tt.args.netMSGs)
		})
	}
}

//
//  Test_sendState
//  @Description: Test_sendState
//  @param t
//
func Test_sendState(t *testing.T) {
	tbftImpl1 := newTbftImpl()
	tbftImpl1.consensusStateCache.getConsensusState(0)
	v, _ := tbftImpl1.GetValidators()
	require.Equal(t, len(v), len(tbftImpl.validatorSet.Validators))
	validatorSet := newValidatorSet(tLogger, nodes, 1)
	heightRoundVoteSet := newHeightRoundVoteSet(
		tLogger, 1, 0, validatorSet)
	//tbftProposal := &TBFTProposal{}
	consensusState := &ConsensusState{tLogger,
		"node1", 1, 1, 1, nil,
		nil, 1, nil, 1,
		nil, heightRoundVoteSet}
	tbftImpl1.logger = tLogger
	tbftImpl1.ConsensusState = consensusState

	var netMSGs []*netpb.NetMsg
	remoter := &RemoteState{}
	status := make(map[int8]consistent_service.Status)
	status[remoter.Type()] = remoter
	r := &Node{id: "node1", status: status}

	type args struct {
		local      *ConsensusTBFTImpl
		remoteInfo consistent_service.Node
		remoter    *RemoteState
		netMSGs    *[]*netpb.NetMsg
	}
	tests := []struct {
		name string
		args args
	}{
		{"test",
			args{tbftImpl1, r,
				remoter,
				&netMSGs}},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			sendState(tt.args.local, tt.args.remoteInfo, tt.args.remoter, tt.args.netMSGs)
		})
	}
}

func TestTbftConsistentMessage_OnMessage(t *testing.T) {
	type fields struct {
		log      consistent_service.Logger
		msgBus   msgbus.MessageBus
		messageC chan interface{}
	}
	type args struct {
		message *msgbus.Message
	}
	tests := []struct {
		name   string
		fields fields
		args   args
	}{
		{"test1",
			fields{newMockLogger(), tbftImpl.msgbus, make(chan interface{})},
			args{&msgbus.Message{Topic: -1, Payload: nil}}},
		//{"test2",
		//	fields{newMockLogger(), tbftImpl.msgbus, make(chan interface{})},
		//	args{&msgbus.Message{Topic: msgbus.RecvConsistentMsg, Payload: nil}}},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			m := &TbftConsistentMessage{
				log:      tt.fields.log,
				msgBus:   tt.fields.msgBus,
				messageC: tt.fields.messageC,
			}
			m.OnMessage(tt.args.message)
		})
	}
}

func TestTbftConsistentMessage_OnQuit(t *testing.T) {
	type fields struct {
		log      consistent_service.Logger
		msgBus   msgbus.MessageBus
		messageC chan interface{}
	}
	tests := []struct {
		name   string
		fields fields
	}{
		{"test1",
			fields{newMockLogger(), tbftImpl.msgbus, make(chan interface{})},
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			m := &TbftConsistentMessage{
				log:      tt.fields.log,
				msgBus:   tt.fields.msgBus,
				messageC: tt.fields.messageC,
			}
			m.OnQuit()
		})
	}
}
